/*
 * This file and its contents are supplied under the terms of the
 * Common Development and Distribution License ("CDDL"), version 1.0.
 * You may only use this file in accordance with the terms of version
 * 1.0 of the CDDL.
 *
 * A full copy of the text of the CDDL should have accompanied this
 * source.  A copy of the CDDL is also available via the Internet at
 * http://www.illumos.org/license/CDDL.
 */
/*
 * Copyright 2012 Nexenta Systems, Inc.  All rights reserved.
 *
 * seqlock.h - Definitions of sequence lock algorithm.
 *
 * Example for reader:
 *
 * do {
 *    int seq = seq_read_begin(&seqlock);
 *    ...
 * } while (seq_read_retry(&seqlock, seq));
 *
 */

#ifndef	_SYS_SEQLOCK_H
#define	_SYS_SEQLOCK_H

#ifdef	__cplusplus
extern "C" {
#endif

#include <sys/types.h>
#include <sys/mutex.h>
#include <sys/atomic.h>		/* for membar_consumer() */

typedef struct seqlock {
	int	 sequence;
	kmutex_t lock;
} seqlock_t;

extern void seq_init(seqlock_t *);
extern void seq_destroy(seqlock_t *);
extern int  seq_read_begin(volatile seqlock_t *);

/*
 * Code under seq_write_lock() should be fast and
 * must not call sleeping function
 */
extern void seq_write_lock(seqlock_t *);
extern void seq_write_unlock(seqlock_t *);

static boolean_t seq_read_retry(const seqlock_t *seqlock, int seq)
{
	membar_consumer();
	return (seqlock->sequence != seq);
}
#pragma inline(seq_read_retry)

#ifdef	__cplusplus
}
#endif

#endif	/* _SYS_SEQLOCK_H */
